#include "model.h"
#include <cassert>
#include <stdexcept>

Player other_player(Player p)
{
    switch (p) {
        case Player::first: return Player::second;
        case Player::second: return Player::first;
        default:
            throw std::invalid_argument("other_player: not a player");
    }
}

Connect4_model::Connect4_model()
        : columns_(m, column_t{})
{ }

const int Connect4_model::k = 4;  // how many to connect
const int Connect4_model::m = 7;  // grid width
const int Connect4_model::n = 6;  // grid height

void Connect4_model::place_token(int col_no)
{
    if (turn_ == Player::neither)
        throw std::logic_error("Model::place_token: game over");

    column_t* col = get_column_(col_no);
    if (col->size() == n)
        throw std::invalid_argument("Model::place_token: column full");

    col->push_back(turn_);
    update_winner_and_turn_(col_no);
}

void Connect4_model::update_winner_and_turn_(int const col_no)
{
    int const row_no = (int)columns_[col_no].size() - 1;

    auto const count = [&](int dcol, int drow) {
        return count_from_by_(turn_, col_no, row_no, dcol, drow);
    };

    int below       = count( 0, -1);
    int left        = count(-1,  0);
    int right       = count( 1,  0);
    int above_left  = count(-1,  1);
    int above_right = count( 1,  1);
    int below_left  = count(-1, -1);
    int below_right = count( 1, -1);

    if (below + 1 >= k ||
            left + 1 + right >= k ||
            above_left + 1 + below_right >= k ||
            above_right + 1 + below_left >= k) {
        winner_ = turn_;
    } else {
        for (column_t const& column : columns_) {
            if (column.size() < n) {
                turn_ = other_player(turn_);
                return;
            }
        }
    }

    turn_ = Player::neither;
}

const Connect4_model::column_t* Connect4_model::get_column(int col_no) const
{
    assert(is_good_col(col_no));
    return &columns_[col_no];
}

Connect4_model::column_t* Connect4_model::get_column_(int col_no)
{
    assert(is_good_col(col_no));
    return &columns_[col_no];
}

int Connect4_model::count_from_by_(Player goal,
                          int start_col, int start_row,
                          int change_col, int change_row) const
{
    int count = 0;

    for (;;) {
        start_col += change_col;
        start_row += change_row;

        if (!is_good_col_row_(start_col, start_row))
            return count;

        if (columns_[start_col][start_row] != goal)
            return count;

        ++count;
    }
}

bool Connect4_model::is_good_col(int col_no) const
{
    return 0 <= col_no && col_no < m;
}

bool Connect4_model::is_playable_col(int col_no) const
{
    return is_good_col(col_no) &&
           get_column(col_no)->size() < Connect4_model::n;
}

bool Connect4_model::is_good_col_row_(int col_no, int row_no) const
{
    return is_good_col(col_no) &&
           0 <= row_no && row_no < columns_[col_no].size();
}
