#pragma once

#include <vector>

/// How we represent players or the absence thereof.
enum class Player
{
    first, second, neither
};

/// Returns the other player, if given Player::first or Player::second;
/// throws std::invalid_argument if given Player::neither.
Player other_player(Player);

/// Models a Connect Four game.
struct Connect4_model
{
    ///
    /// A TYPE ABBREVIATION:
    ///

    // A type member used below - we represent a column as a vector of
    // `Player`s:
    using column_t = std::vector<Player>;


    ///
    /// CONSTRUCTOR
    ///

    /// Constructs an empty Connect Four game model.
    Connect4_model();


    ///
    /// CONSTANTS
    ///

    // The game parameters.
    static const int k;  // how many to connect (4)
    static const int m;  // grid width (7)
    static const int n;  // grid height (6)


    ///
    /// API FUNCTIONS (PUBLIC MEMBER FUNCTIONS)
    ///

    /// Gets a view of the given column.
    ///
    /// **PRECONDITION:** `0 <= col_no < m()`
    const column_t* get_column(int col_no) const;

    /// Places the token for the given player in the given column, provided
    /// that column is not full.
    ///
    /// **PRECONDITIONS**:
    ///  - `0 <= col_no < m()`
    ///  - `get_column(col_no).size() < n()`
    void place_token(int col_no);

    /// Is the column number within bounds?
    bool is_good_col(int col_no) const;

    /// Is there room to play in the given column?
    bool is_playable_col(int col_no) const;

    /// Returns whose turn it is, or Player::neither for game over.
    Player get_turn() const { return turn_; };

    /// Returns the winner, or Player::neither for stalemates or when
    /// the game is not yet over.
    Player get_winner() const { return winner_; };

    ///
    /// INTERNAL FUNCTIONS (PRIVATE MEMBER FUNCTIONS)
    ///

    /// Gets a mutable view of the given column.
    ///
    /// **PRECONDITION:** `col_no < m()`
    column_t* get_column_(int col_no);

    /// **PRECONDITION** `col_no` is valid and where the last move
    /// was played.
    void update_winner_and_turn_(int col_no);

    /// Counts the number of instances of `goal`, moving in direction
    /// {change_col, change_row} starting at {start_col, start_row}, and
    /// not counting the Player at the starting position.
    int count_from_by_(Player goal,
                       int start_col, int start_row,
                       int change_col, int change_row) const;

    /// Returns whether there is a token at the given position.
    bool is_good_col_row_(int col_no, int row_no) const;

    ///
    /// FIELDS (PRIVATE DATA MEMBERS)
    ///

    // The current turn.
    Player turn_   = Player::first;

    // The winning player, if any.
    Player winner_ = Player::neither;

    // The grid itself.
    std::vector<column_t> columns_;
    // INVARIANT:
    //  - columns_.size() == m
    //  - for (column_t const& column : columns) column.size() <= n
    //  - none of the Player values is Player::neither
};

